"""Home Assistant Device Module.

This module provides the base device class for interacting with Home Assistant devices.
It defines common functionality for controlling devices and getting their state information.
"""

from ovos_utils.log import LOG
from webcolors import (  # TODO: Use ovos-color-parser when it's ready
    name_to_rgb,
    rgb_to_name,
)

from skill_homeassistant.ha_client.logic.base import HomeAssistantConnector


class HomeAssistantDevice:
    """Home Assistant Device"""

    def __init__(  # pylint: disable=keyword-arg-before-vararg
        self,
        connector: HomeAssistantConnector,
        device_id,
        device_icon,
        device_name,
        device_state,
        device_attributes,
        device_area=None,
    ):
        """Initialize the device.

        Args:
            connector (HomeAssistantConnector): The connector to use.
            device_id (str): The id of the device.
            device_icon (str): The icon of the device.
            device_name (str): The name of the device.
            device_state (str): The state of the device.
            device_attributes (dict): The attributes of the device.
            device_area (str): Name of area device is in (None if no area)
        """
        self.connector = connector
        self.device_id = device_id
        self.device_icon = device_icon
        self.device_name = device_name
        self.device_state = device_state
        self.device_attributes = device_attributes
        self.device_area = device_area
        self.has_device_class = False
        self.device_class = None
        self.device_type = self.device_id.split(".")[0]
        self.query_device_class()
        self.connector.register_callback(self.device_id, self.callback_listener)

    def callback_listener(self, message):
        """Callback for when the device state changes."""
        event = message.get("event")
        event_type = event.get("event_type")
        if event_type == "state_changed":
            new_state = event.get("data").get("new_state")
            if new_state.get("entity_id") == self.device_id:
                self.device_state = new_state.get("state")
                self.device_attributes = new_state.get("attributes")

    def query_device_class(self):
        """Query the device class of the device."""
        if "device_class" in self.device_attributes:
            self.has_device_class = True
            self.device_class = self.device_attributes["device_class"]
        else:
            self.has_device_class = False
            self.device_class = None

    def get_has_device_class(self):
        """Get if the device has a device class."""
        return self.has_device_class

    def get_device_class(self):
        """Get the device class of the device."""
        return self.device_class

    def get_state(self):
        """Get the state of the device."""
        return self.device_state

    def get_state_json_object(self):
        """Get the state of the device as a json object."""
        return self.connector.get_device_state(self.device_id)

    def get_attribute(self, attribute):
        """Get an attribute of the device."""
        return self.device_attributes[attribute]

    def get_attributes(self):
        """Get the attributes of the device."""
        return self.device_attributes

    def get_id(self):
        """Get the id of the device."""
        return self.device_id

    def get_icon(self):
        """Get the icon of the device."""
        return self.device_icon

    def get_name(self):
        """Get the name of the device."""
        return self.device_name

    def get_supported_features(self):
        """Get the supported features of the device."""
        return self.device_attributes["supported_features"]

    def is_on(self):
        """Check if the device is on."""
        return self.device_state == "on"

    def is_off(self):
        """Check if the device is off."""
        return self.device_state == "off"

    def is_unavailable(self):
        """Check if the device is unavailable."""
        return self.device_state == "unavailable"

    def turn_on(self):
        """Turn on the device."""
        return self.connector.turn_on(self.device_id, self.device_type)

    def turn_off(self):
        """Turn off the device."""
        return self.connector.turn_off(self.device_id, self.device_type)

    def call_function(self, function_name, function_args=None):
        """Call a function of the device.

        Args:
            function_name (str): The name of the function to call.
            function_args (dict): The arguments to pass to the function.
        """
        return self.connector.call_function(self.device_id, self.device_type, function_name, function_args)

    def update_device(self):
        """Update the device."""
        device = self.connector.get_device_state(self.device_id)
        self.device_state = device["state"]
        self.device_attributes = device["attributes"]
        self.device_icon = device["attributes"].get("icon", "")
        self.device_name = device["attributes"].get("friendly_name", "")

    def set_device_attribute(self, device_id, attribute, value):
        """Set an attribute of the device.

        Args:
            device_id (str): The id of the device.
            attribute (str): The attribute to set.
            value (str): The value to set the attribute to.
        """
        device = self.get_state_json_object()
        attributes = device["attributes"]
        attributes[attribute] = value
        self.connector.set_device_state(device_id, device["state"], attributes)

    def poll(self):
        """Poll the device."""
        full_state_json = self.connector.get_device_state(self.device_id)
        if full_state_json:
            if full_state_json == "unavailable":
                LOG.warning(f"State unavailable for device: {self.device_id}")
            elif not isinstance(full_state_json, dict):
                LOG.error(f"({self.device_name}) Expected dict state but got: " f"{full_state_json}")
            else:
                self.device_state = full_state_json.get("state", "unknown")
                self.device_attributes = full_state_json.get("attributes", {})

    def get_device_display_model(self):
        """Get the display model of the device."""
        self.poll()
        return {
            "id": self.device_id,
            "name": self.device_name,
            "icon": self.device_icon,
            "state": self.device_state,
            "type": self.device_type,
            "attributes": self.device_attributes,
            "host": self.connector.host,
        }


class HomeAssistantLight(HomeAssistantDevice):
    """Home Assistant Light"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_brightness(self):
        """Get the brightness of the light."""
        return self.device_attributes.get("brightness", 0)

    def get_color_mode(self):
        """Get the color mode of the light."""
        return self.device_attributes.get("color_mode", "unknown")

    def get_color_temp(self):
        """Get the color temperature of the light."""
        return self.device_attributes.get("color_temp", 0)

    def get_effect(self):
        """Get the effect of the light."""
        return self.device_attributes.get("effect", "none")

    def get_effect_list(self):
        """Get the effect list of the light."""
        return self.device_attributes.get("effect_list", [])

    def get_hs_color(self):
        """Get the hs color of the light."""
        return self.device_attributes.get("hs_color", [0, 0])

    def get_max_mireds(self):
        """Get the max mireds of the light."""
        return self.device_attributes.get("max_mireds", 0)

    def get_min_mireds(self):
        """Get the min mireds of the light."""
        return self.device_attributes.get("min_mireds", 0)

    def get_rgb_color(self):
        """Get the rgb color of the light."""
        return self.device_attributes.get("rgb_color", [0, 0, 0])

    def get_spoken_color(self):
        """Get the spoken color value of the light."""
        color = tuple(self.get_rgb_color())
        try:
            color = rgb_to_name(color)
        except ValueError:
            color = f"RGB code {color[0]}, {color[1]}, {color[2]}"
        return color

    def get_supported_color_modes(self):
        """Get the supported color modes of the light."""
        return self.device_attributes.get("supported_color_modes", [])

    def get_xy_color(self):
        """Get the xy color of the light."""
        return self.device_attributes.get("xy_color", [0, 0])

    def set_brightness(self, brightness):
        """Set the brightness of the light.

        Args:
            brightness (int): The brightness to set the light to.
        """
        LOG.debug(f"Setting brightness to {brightness}")
        self.call_function("turn_on", {"brightness": brightness})
        self.update_device()

    def increase_brightness(self, brightness_increment: int = 10) -> int:
        """Increase the brightness of the light by the brightness increment."""
        bumped_value = self.call_function("turn_on", {"brightness_step_pct": brightness_increment})
        self.update_device()
        return bumped_value

    def decrease_brightness(self, brightness_increment: int = 10) -> int:
        """Decrease the brightness of the light by the brightness increment."""
        decreased_value = self.call_function("turn_on", {"brightness_step_pct": -brightness_increment})
        self.update_device()
        return decreased_value

    def set_color(self, color):
        """Set the color of the light.

        Args:
            color (str): The color to set the light to.
        """
        rgb = name_to_rgb(color)
        LOG.debug(f"Setting color to [{rgb.red}, {rgb.green}, {rgb.blue}]")
        self.set_rgb_color([rgb.red, rgb.green, rgb.blue])
        self.update_device()

    def set_color_mode(self, color_mode):
        """Set the color mode of the light.

        Args:
            color_mode (str): The color mode to set the light to.
        """
        self.call_function("set_color_mode", {"color_mode": color_mode})
        self.update_device()

    def set_color_temp(self, color_temp):
        """Set the color temperature of the light.

        Args:
            color_temp (int): The color temperature to set the light to.
        """
        self.call_function("set_color_temp", {"color_temp": color_temp})
        self.update_device()

    def set_effect(self, effect):
        """Set the effect of the light.

        Args:
            effect (str): The effect to set the light to.
        """
        self.call_function("set_effect", {"effect": effect})
        self.update_device()

    def set_hs_color(self, hs_color):
        """Set the hs color of the light.

        Args:
            hs_color (list): The hs color to set the light to.
        """
        self.call_function("set_hs_color", {"hs_color": hs_color})
        self.update_device()

    def set_rgb_color(self, rgb_color):
        """Set the rgb color of the light.

        Args:
            rgb_color (list): The rgb color to set the light to.
        """
        self.call_function("turn_on", {"rgb_color": rgb_color})
        self.update_device()

    def set_xy_color(self, xy_color):
        """Set the xy color of the light.

        Args:
            xy_color (list): The xy color to set the light to.
        """
        self.call_function("set_xy_color", {"xy_color": xy_color})
        self.update_device()


class HomeAssistantSwitch(HomeAssistantDevice):
    """Home Assistant Switch"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class HomeAssistantSensor(HomeAssistantDevice):
    """Home Assistant Sensor"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_device_class(self):
        """Get the device class of the sensor."""
        return self.device_attributes.get("device_class", "unknown")

    def get_last_reset(self):
        """Get the last reset of the sensor."""
        return self.device_attributes.get("last_reset", "unknown")

    def get_native_value(self):
        """Get the native value of the sensor."""
        return self.device_attributes.get("native_value", "unknown")

    def get_native_unit_of_measurement(self):
        """Get the native unit of measurement of the sensor."""
        return self.device_attributes.get("native_unit_of_measurement", "unknown")

    def get_state_class(self):
        """Get the state class of the sensor."""
        return self.device_attributes.get("state_class", "unknown")

    def get_suggested_unit_of_measurement(self):
        """Get the suggested unit of measurement of the sensor."""
        return self.device_attributes.get("suggested_unit_of_measurement", "unknown")


class HomeAssistantBinarySensor(HomeAssistantDevice):
    """Home Assistant Binary Sensor"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_device_class(self):
        """Get the device class of the binary sensor."""
        return self.device_attributes.get("device_class", "unknown")


class HomeAssistantCover(HomeAssistantDevice):
    """Home Assistant Cover"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def open(self):
        """Open the cover."""
        self.call_function("open")

    def close(self):
        """Close the cover."""
        self.call_function("close")

    def set_position(self, position):
        """Set the position of the cover.

        Args:
            position (int): The position to set the cover to.
        """
        self.call_function("set_position", {"position": position})
        self.update_device()

    def stop(self):
        """Stop the cover."""
        self.call_function("stop")

    def is_opening(self):
        """Check if the cover is opening."""
        return self.device_state == "opening"

    def is_closing(self):
        """Check if the cover is closing."""
        return self.device_state == "closing"

    def is_open(self):
        """Check if the cover is open."""
        return self.device_state == "open"

    def is_closed(self):
        """Check if the cover is closed."""
        return self.device_state == "closed"

    def get_position(self):
        """Get the position of the cover."""
        return self.device_attributes["current_position"]


class HomeAssistantMediaPlayer(HomeAssistantDevice):
    """Home Assistant Media Player"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_media_title(self):
        """Get the media title of the media player."""
        return self.device_attributes["media_title"]

    def get_media_artist(self):
        """Get the media artist of the media player."""
        return self.device_attributes["media_artist"]

    def get_media_album_name(self):
        """Get the media album name of the media player."""
        return self.device_attributes["media_album_name"]

    def get_media_series_title(self):
        """Get the media series title of the media player."""
        return self.device_attributes["media_series_title"]

    def get_media_season(self):
        """Get the media season of the media player."""
        return self.device_attributes["media_season"]

    def get_media_episode(self):
        """Get the media episode of the media player."""
        return self.device_attributes["media_episode"]

    def get_media_channel(self):
        """Get the media channel of the media player."""
        return self.device_attributes["media_channel"]

    def get_media_content_id(self):
        """Get the media content id of the media player."""
        return self.device_attributes["media_content_id"]

    def get_media_content_type(self):
        """Get the media content type of the media player."""
        return self.device_attributes["media_content_type"]

    def get_media_duration(self):
        """Get the media duration of the media player."""
        return self.device_attributes["media_duration"]

    def get_media_position(self):
        """Get the media position of the media player."""
        return self.device_attributes["media_position"]

    def get_media_position_updated_at(self):
        """Get the media position updated at of the media player."""
        return self.device_attributes["media_position_updated_at"]

    def get_is_volume_muted(self):
        """Get the is volume muted of the media player."""
        return self.device_attributes["is_volume_muted"]

    def get_volume_level(self):
        """Get the volume level of the media player."""
        return self.device_attributes["volume_level"]

    def get_app_id(self):
        """Get the app id of the media player."""
        return self.device_attributes["app_id"]

    def get_app_name(self):
        """Get the app name of the media player."""
        return self.device_attributes["app_name"]


class HomeAssistantClimate(HomeAssistantDevice):
    """Home Assistant Climate"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def set_temperature(self, temperature):
        """Set the temperature of the climate device.

        Args:
            temperature (float): The temperature to set the climate device to.
        """
        self.call_function("set_temperature", {"temperature": temperature})
        self.update_device()

    def set_hvac_mode(self, hvac_mode):
        """Set the hvac mode of the climate device.

        Args:
            hvac_mode (str): The hvac mode to set the climate device to.
        """
        self.call_function("set_hvac_mode", {"hvac_mode": hvac_mode})
        self.update_device()

    def set_fan_mode(self, fan_mode):
        """Set the fan mode of the climate device.

        Args:
            fan_mode (str): The fan mode to set the climate device to.
        """
        self.call_function("set_fan_mode", {"fan_mode": fan_mode})
        self.update_device()

    def set_swing_mode(self, swing_mode):
        """Set the swing mode of the climate device.

        Args:
            swing_mode (str): The swing mode to set the climate device to.
        """
        self.call_function("set_swing_mode", {"swing_mode": swing_mode})
        self.update_device()

    def set_preset_mode(self, preset_mode):
        """Set the preset mode of the climate device.

        Args:
            preset_mode (str): The preset mode to set the climate device to.
        """
        self.call_function("set_preset_mode", {"preset_mode": preset_mode})
        self.update_device()

    def set_aux_heat(self, aux_heat):
        """Set the aux heat of the climate device.

        Args:
            aux_heat (bool): The aux heat to set the climate device to.
        """
        self.call_function("set_aux_heat", {"aux_heat": aux_heat})
        self.update_device()

    def set_humidity(self, humidity):
        """Set the humidity of the climate device.

        Args:
            humidity (float): The humidity to set the climate device to.
        """
        self.call_function("set_humidity", {"humidity": humidity})
        self.update_device()

    def set_target_humidity(self, target_humidity):
        """ Set the target humidity of the climate device.\

        Args:
            target_humidity (float): The target humidity to set the climate device to.
        """
        self.call_function("set_target_humidity", {"target_humidity": target_humidity})
        self.update_device()

    def set_target_temp_low(self, target_temp_low):
        """Set the target temp low of the climate device.

        Args:
            target_temp_low (float): The target temp low to set the climate device to.
        """
        self.call_function("set_target_temp_low", {"target_temp_low": target_temp_low})
        self.update_device()

    def set_target_temp_high(self, target_temp_high):
        """Set the target temp high of the climate device.

        Args:
            target_temp_high (float): The target temp high to set the climate device to.
        """
        self.call_function("set_target_temp_high", {"target_temp_high": target_temp_high})
        self.update_device()

    def get_current_temperature(self):
        """Get the current temperature of the climate device."""
        return self.device_attributes["current_temperature"]

    def get_current_humidity(self):
        """Get the current humidity of the climate device."""
        return self.device_attributes["current_humidity"]

    def get_temperature(self):
        """Get the temperature of the climate device."""
        return self.device_attributes["temperature"]

    def get_target_temp_low(self):
        """Get the target temp low of the climate device."""
        return self.device_attributes["target_temp_low"]

    def get_target_temp_high(self):
        """Get the target temp high of the climate device."""
        return self.device_attributes["target_temp_high"]

    def get_humidity(self):
        """Get the humidity of the climate device."""
        return self.device_attributes["humidity"]

    def get_target_humidity(self):
        """Get the target humidity of the climate device."""
        return self.device_attributes["target_humidity"]

    def get_min_temp(self):
        """Get the min temp of the climate device."""
        return self.device_attributes["min_temp"]

    def get_max_temp(self):
        """Get the max temp of the climate device."""
        return self.device_attributes["max_temp"]

    def get_target_temp_step(self):
        """Get the target temp step of the climate device."""
        return self.device_attributes["target_temp_step"]

    def get_hvac_mode(self):
        """Get the hvac mode of the climate device."""
        return self.device_attributes["hvac_mode"]

    def get_hvac_modes(self):
        """Get the hvac modes of the climate device."""
        return self.device_attributes["hvac_modes"]

    def get_fan_mode(self):
        """Get the fan mode of the climate device."""
        return self.device_attributes["fan_mode"]


class HomeAssistantVacuum(HomeAssistantDevice):
    """Home Assistant Vacuum"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def start(self):
        """Start the vacuum."""
        self.call_function("start")

    def pause(self):
        """Pause the vacuum."""
        self.call_function("pause")

    def stop(self):
        """Stop the vacuum."""
        self.call_function("stop")

    def return_to_base(self):
        """Return the vacuum to base."""
        self.call_function("return_to_base")

    def set_fan_speed(self, fan_speed):
        """Set the fan speed of the vacuum.

        Args:
            fan_speed (str): The fan speed to set the vacuum to.
        """
        self.set_device_attribute(self.device_id, "fan_speed", fan_speed)
        self.update_device()

    def send_command(self, command, params):
        """Send a command to the vacuum.

        Args:
            command (str): The command to send to the vacuum.
            params (dict): The parameters to send to the vacuum.
        """
        self.call_function("send_command", {"command": command, "params": params})

    def get_battery_level(self):
        """Get the battery level of the vacuum."""
        return self.device_attributes["battery_level"]

    def get_fan_speed(self):
        """Get the fan speed of the vacuum."""
        return self.device_attributes["fan_speed"]

    def get_fan_speed_list(self):
        """Get the fan speed list of the vacuum."""
        return self.device_attributes["fan_speed_list"]

    def get_status(self):
        """Get the status of the vacuum."""
        return self.device_attributes["status"]


class HomeAssistantCamera(HomeAssistantDevice):
    """Home Assistant Camera"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class HomeAssistantScene(HomeAssistantDevice):
    """Home Assistant Scene"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def turn_off(self):
        LOG.warning("Request to turn off a scene. This is not supported - scenes can only be activated.")
        return


class HomeAssistantAutomation(HomeAssistantDevice):
    """Home Assistant Automation"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def turn_off(self):
        LOG.warning("Request to turn off an automation. This is not supported, as it will disable it instead.")
        return



# pylint: disable=missing-function-docstring,missing-class-docstring,missing-module-docstring,logging-fstring-interpolation
from typing import Optional

from ovos_bus_client import Message
from ovos_workshop.decorators import intent_handler
from ovos_workshop.skills import OVOSSkill

from skill_homeassistant.ha_client import HomeAssistantClient


class HomeAssistantSkill(OVOSSkill):
    """Unified Home Assistant skill for OpenVoiceOS or Neon.AI."""

    _settings_defaults = {
        "host": "",
        "api_key": "",
        "silent_entities": set(),
        "disable_intents": False,
        "timeout": 5,
        "verify_ssl": True,
    }
    _intents_enabled = True
    connected_intents = (
        "sensor.intent",
        "turn.on.intent",
        "turn.off.intent",
        "stop.intent",
        "lights.get.brightness.intent",
        "lights.set.brightness.intent",
        "lights.increase.brightness.intent",
        "lights.decrease.brightness.intent",
        "lights.get.color.intent",
        "lights.set.color.intent",
        "assist.intent",
    )

    def __init__(self, *args, bus=None, skill_id="", **kwargs):
        super().__init__(*args, bus=bus, skill_id=skill_id, **kwargs)

    @property
    def verify_ssl(self):
        """Return whether to verify SSL connections."""
        return self._get_setting("verify_ssl")

    @property
    def silent_entities(self):
        return set(self._get_setting("silent_entities"))

    @silent_entities.setter
    def silent_entities(self, value):
        self._set_setting("silent_entities", value)

    @property
    def disable_intents(self):
        setting = self._get_setting("disable_intents")
        self._handle_connection_state(setting)
        return setting

    @disable_intents.setter
    def disable_intents(self, value):
        self._set_setting("disable_intents", value)
        self._handle_connection_state(value)

    def initialize(self):
        self.client_config = self._get_client_config()  # pylint: disable=attribute-defined-outside-init
        self.ha_client = HomeAssistantClient(  # pylint: disable=attribute-defined-outside-init
            config=self.client_config, bus=self.bus
        )
        if self.disable_intents:
            self.log.info("User has indicated they do not want to use Home Assistant intents. Disabling.")
            self.disable_ha_intents()

        # Register for settings changes to update client config
        self.settings_change_callback = self._on_settings_changed

    def _on_settings_changed(self):
        """Handle settings changes by updating the Home Assistant client config."""
        self.log.info("Settings changed, updating Home Assistant client configuration")
        new_config = self._get_client_config()
        self.ha_client.update_config(new_config)

    def _get_client_config(self) -> dict:
        if self.settings.get("host") and self.settings.get("api_key"):
            return {**self._settings_defaults, **self.settings}
        phal_config = self.config_core.get("PHAL", {}).get("ovos-PHAL-plugin-homeassistant")
        if phal_config:
            return {**self._settings_defaults, **phal_config, **self.settings}
        self.log.error(
            "No Home Assistant config found! Please set host and api_key "
            f"in the skill settings at {self.settings_path}."
        )
        return self._settings_defaults

    def _get_setting(self, setting_name):
        """Helper method to get a setting with its default value."""
        return self.settings.get(setting_name, self._settings_defaults[setting_name])

    def _set_setting(self, setting_name, value):
        """Helper method to set a setting."""
        self.settings[setting_name] = value

    def _handle_connection_state(self, disable_intents: bool):
        if self._intents_enabled and disable_intents is True:
            self.log.info(
                "Disabling Home Assistant intents by user request. To re-enable, set disable_intents to False."
            )
            self.disable_ha_intents()
        if not self._intents_enabled and disable_intents is False:
            self.log.info("Enabling Home Assistant intents by user request. To disable, set disable_intents to True.")
            self.enable_ha_intents()

    def enable_ha_intents(self):
        for intent in self.connected_intents:
            success = self.enable_intent(intent)
            if not success:
                self.log.error(f"Error registering intent: {intent}")
            else:
                self.log.info(f"Successfully registered intent: {intent}")
        self._intents_enabled = True

    def disable_ha_intents(self):
        for intent in self.connected_intents:
            self.intent_service.remove_intent(intent)
            try:
                assert self.intent_service.intent_is_detached(intent) is True
            except AssertionError:
                self.log.error(f"Error disabling intent: {intent}")
        self._intents_enabled = False

    # Handlers
    @intent_handler("get.all.devices.intent")
    def handle_rebuild_device_list(self, _: Message):
        if not self.check_client_connection():
            self.log.warning("Cannot rebuild device list: Home Assistant connection not available")
            return
        self.gui.show_text("Refreshing device list from Home Assistant...")
        device_count = self.ha_client.refresh_devices()
        self.speak_dialog("rebuild.complete", data={"count": device_count})
        self.gui.show_text(f"Device list refreshed: {device_count} devices found")

    @intent_handler("enable.intent")
    def handle_enable_intent(self, _: Message):
        self.settings["disable_intents"] = False
        self.speak_dialog("enable")
        self.enable_ha_intents()

    @intent_handler("disable.intent")
    def handle_disable_intent(self, _: Message):
        self.settings["disable_intents"] = True
        self.speak_dialog("disable")
        self.disable_ha_intents()

    @intent_handler("sensor.intent")  # pragma: no cover
    def get_device_intent(self, message: Message):
        """Handle intent to get a single device status from Home Assistant."""
        self.log.info(message.data)
        if not self.check_client_connection():
            return
        device = message.data.get("entity", "")
        if device:
            device_data = self.ha_client.handle_get_device(Message("", {"device": device}))
            if device_data:
                device_name = device_data.get("attributes", {}).get("friendly_name", device_data.get("name"))
                device_type = device_data.get("type")
                device_state = device_data.get("state")
                self.speak_dialog(
                    "device.status",
                    data={
                        "device": device_name,
                        "type": device_type,
                        "state": device_state,
                    },
                )
                self.gui.show_text(f"{device_name} ({device_type}) is {device_state}")
            else:
                self.speak_dialog("device.not.found", {"device": device})
                self.gui.show_text(f"Could not find device {device}")
            self.log.info(f"Trying to get device status for {device}")
        else:
            self.speak_dialog("no.parsed.device")

    def _get_device_from_message(self, message: Message, require_device: bool = True) -> Optional[str]:
        """Extract and validate device from message data.

        Args:
            message: The message containing device data
            require_device: If True, speak no.parsed.device dialog when device is missing

        Returns:
            The device name or None if not found/invalid
        """
        device = message.data.get("entity", "")
        if not device and require_device:
            self.speak_dialog("no.parsed.device")
            return None
        return device or None

    def _handle_device_response(
        self,
        response: Optional[dict],
        device: str,
        success_dialog: str,
        success_data: Optional[dict] = None,
        success_message: str = "Successful operation!",
    ) -> bool:
        """Handle standard device operation response.

        Args:
            response: The response from ha_client
            device: The device name
            success_dialog: Dialog to speak on success
            success_data: Additional data to pass to success dialog

        Returns:
            True if handled successfully, False otherwise
        """
        if not response or response.get("response"):
            self.speak_dialog("device.not.found", {"device": device})
            self.gui.show_text(f"Could not find device {device}")
            return False

        if device not in self.silent_entities:
            dialog_data = {"device": device}
            if success_data:
                dialog_data.update(success_data)
            self.speak_dialog(success_dialog, dialog_data)

        self.gui.show_text(f"{device}: {success_message}")
        return True
    
    def check_client_connection(self):
        if not self.ha_client.instance_available:
            try:
                self.ha_client.init_configuration(self.settings)
                if not self.ha_client.instance_available:
                    raise Exception("Home Assistant instance is not available")
                return True
            except Exception as e:
                self.log.error(f"Error initializing Home Assistant client: {e}")
                self.gui.show_text("Connection to Home Assistant is not configured or unavailable. Please check skill settings.")
                self.speak_dialog("device.status", data={
                    "device": "Home Assistant",
                    "type": "server",
                    "state": "not configured, check your skill settings or connection to Home Assistant instance."
                    })
                return False
        return True

    @intent_handler("turn.on.intent")  # pragma: no cover
    def handle_turn_on_intent(self, message: Message) -> None:
        """Handle turn on intent."""
        self.log.info(message.data)
        if not self.check_client_connection():
            return
        if device := self._get_device_from_message(message):
            response = self.ha_client.handle_turn_on(Message("", {"device": device}))
            if not self._handle_device_response(
                response, device, "device.turned.on", success_message="Successfully turned on!"
            ):
                self.log.info(f"Trying to turn on device {device}")

    @intent_handler("turn.off.intent")  # pragma: no cover
    @intent_handler("stop.intent")  # pragma: no cover
    def handle_turn_off_intent(self, message: Message) -> None:
        """Handle turn off intent."""
        self.log.info(message.data)
        if not self.check_client_connection():
            return
        if device := self._get_device_from_message(message):
            response = self.ha_client.handle_turn_off(Message("", {"device": device}))
            if not self._handle_device_response(
                response, device, "device.turned.off", success_message="Successfully turned off/stopped!"
            ):
                self.log.info(f"Trying to turn off device {device}")

    @intent_handler("lights.get.brightness.intent")  # pragma: no cover
    def handle_get_brightness_intent(self, message: Message):
        self.log.info(message.data)
        if not self.check_client_connection():
            return
        if device := self._get_device_from_message(message):
            response = self.ha_client.handle_get_light_brightness(Message("", {"device": device}))
            if response and not response.get("response"):
                if brightness := response.get("brightness"):
                    self.speak_dialog(
                        "lights.current.brightness",
                        data={"brightness": brightness, "device": device},
                    )
                    self.gui.show_text(f"{device}: Current brightness {brightness}")
                    return
            self.speak_dialog("lights.status.not.available", data={"device": device})
            self.gui.show_text(f"{device} not found in Home Assistant.")

    @intent_handler("lights.set.brightness.intent")  # pragma: no cover
    def handle_set_brightness_intent(self, message: Message):
        self.log.info(message.data)
        if not self.check_client_connection():
            return
        device = self._get_device_from_message(message)
        brightness = message.data.get("brightness")

        if device and brightness:
            response = self.ha_client.handle_set_light_brightness(
                Message(
                    "", {"device": device, "brightness": self._get_ha_value_from_percentage_brightness(brightness)}
                )
            )
            if self._handle_device_response(
                response,
                device,
                "lights.current.brightness",
                {"brightness": response.get("brightness")} if response else None,
                success_message=f"Brightness set to {brightness}",
            ):
                return
            self.log.info(f"Trying to set brightness of {brightness} for {device}")

    @intent_handler("lights.increase.brightness.intent")  # pragma: no cover
    def handle_increase_brightness_intent(self, message: Message):
        self.log.info(message.data)
        if not self.check_client_connection():
            return
        if device := self._get_device_from_message(message):
            response = self.ha_client.handle_increase_light_brightness(Message("", {"device": device}))
            brightness = response.get("brightness", "unknown percentage")
            if self._handle_device_response(
                response,
                device,
                "lights.current.brightness",
                {"brightness": brightness} if response else None,
                success_message=f"Increased brightness to {brightness}",
            ):
                return
            self.log.info(f"Trying to increase brightness for {device}")

    @intent_handler("lights.decrease.brightness.intent")  # pragma: no cover
    def handle_decrease_brightness_intent(self, message: Message):
        self.log.info(message.data)
        if not self.check_client_connection():
            return
        if device := self._get_device_from_message(message):
            response = self.ha_client.handle_decrease_light_brightness(Message("", {"device": device}))
            brightness = response.get("brightness", "unknown percentage")
            if self._handle_device_response(
                response,
                device,
                "lights.current.brightness",
                {"brightness": brightness},
                success_message=f"Decreased brightness to {brightness}",
            ):
                return
            self.log.info(f"Trying to decrease brightness for {device}")

    @intent_handler("lights.get.color.intent")  # pragma: no cover
    def handle_get_color_intent(self, message: Message):
        self.log.info(message.data)
        if not self.check_client_connection():
            return
        if device := self._get_device_from_message(message):
            response = self.ha_client.handle_get_light_color(Message("", {"device": device}))
            if response and not response.get("response"):
                if color := response.get("color"):
                    self.speak_dialog("lights.current.color", data={"color": color, "device": device})
                    self.gui.show_text(f"{device}: Current color {color}")
                    return
            self.speak_dialog("lights.status.not.available", data={"device": device})
            self.gui.show_text(f"Could not get color of {device}")

    @intent_handler("lights.set.color.intent")  # pragma: no cover
    def handle_set_color_intent(self, message: Message):
        self.log.info(message.data)
        if not self.check_client_connection():
            return
        device = self._get_device_from_message(message)
        color = message.data.get("color")

        if not color:
            self.speak_dialog("no.parsed.color")
            return

        color = color.strip()
        if color.startswith("to "):
            color = color[3:].strip()

        if device:
            response = self.ha_client.handle_set_light_color(Message("", {"device": device, "color": color}))
            color = response.get("color", "unknown")
            if self._handle_device_response(
                response,
                device,
                "lights.current.color",
                {"color": color},
                success_message=f"Set color to {color}",
            ):
                return
            self.log.info(f"Trying to set color of {device}")

    @intent_handler("assist.intent")  # pragma: no cover
    def handle_assist_intent(self, message: Message):
        """Handle passthrough to Home Assistant's Assist API."""
        if not self.check_client_connection():
            return
        command = message.data.get("command")
        if command:
            self.ha_client.handle_assist_message(Message("", {"command": command}))
            self.speak_dialog("assist")
            self.log.info(f"Trying to pass message to Home Assistant's Assist API:\n{command}")
            self.gui.show_text(f"Sending message to Assist: {command}")
        else:
            self.speak_dialog("assist.not.understood")

    def _get_ha_value_from_percentage_brightness(self, brightness):
        return round(int(brightness)) / 100 * 255



"""Home Assistant REST Connector Module.

This module provides the REST API connector implementation for interacting with
Home Assistant. It makes HTTP calls to control devices and get state information.
"""

import json

import requests
from ovos_utils.log import LOG

from skill_homeassistant.ha_client.logic.base import HomeAssistantConnector


class HomeAssistantRESTConnector(HomeAssistantConnector):
    """Home Assistant REST Connector"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.headers = {
            "Authorization": "Bearer " + self.api_key,
            "content-type": "application/json",
        }

    def register_callback(self, device_id, callback):
        self.event_listeners[device_id] = callback

    def get_all_devices(self):
        """Get all devices from home assistant."""
        url = self.host + "/api/states"
        try:
            response = requests.get(url, headers=self.headers, timeout=self.timeout, verify=self.verify_ssl)
            response.raise_for_status()
            return response.json()
        except requests.exceptions.ConnectionError:
            LOG.exception(f"Error connecting to Home Assistant at {self.host}")
            return []
        except requests.exceptions.RequestException:
            LOG.exception("Error fetching devices")
            return []

    def get_device_state(self, entity_id):
        """Get the state of a device."""
        url = self.host + "/api/states/" + entity_id
        try:
            response = requests.get(url, headers=self.headers, timeout=self.timeout, verify=self.verify_ssl)
            response.raise_for_status()
            return response.json()
        except requests.exceptions.ConnectionError:
            LOG.exception(f"Error connecting to Home Assistant at {self.host}")
            return []
        except requests.exceptions.RequestException:
            LOG.exception("Error fetching device state")
            return {}

    def set_device_state(self, entity_id, state, attributes=None):
        """Set the state of a device.

        Args:
            entity_id (str): The id of the device.
            state (str): The state to set.
            attributes (dict): The attributes to set.
        """
        url = self.host + "/api/states/" + entity_id
        payload = {"state": state, "attributes": attributes}
        response = requests.post(
            url, data=json.dumps(payload), headers=self.headers, timeout=self.timeout, verify=self.verify_ssl
        )
        try:
            response.raise_for_status()
            return response.json()
        except requests.exceptions.RequestException:
            LOG.exception("Error setting device state")
            return None

    def get_all_devices_with_type(self, device_type):
        """Get all devices with a specific type.

        Args:
            device_type (str): The type of the device.
        """
        devices = self.get_all_devices()
        return [device for device in devices if device["entity_id"].startswith(device_type)]

    def get_all_devices_with_type_and_attribute(self, device_type, attribute, value):
        """Get all devices with a specific type and attribute.

        Args:
            device_type (str): The type of the device.
            attribute (str): The attribute to check.
            value (str): The value of the attribute.
        """
        devices = self.get_all_devices()
        return [
            device
            for device in devices
            if device["entity_id"].startswith(device_type) and device["attributes"][attribute] == value
        ]

    def get_all_devices_with_type_and_attribute_in(self, device_type, attribute, value):
        """Get all devices with a specific type and attribute.

        Args:
            device_type (str): The type of the device.
            attribute (str): The attribute to check.
            value (str): The value of the attribute.
        """
        devices = self.get_all_devices()
        return [
            device
            for device in devices
            if device["entity_id"].startswith(device_type) and device["attributes"][attribute] in value
        ]

    def get_all_devices_with_type_and_attribute_not_in(self, device_type, attribute, value):
        """Get all devices with a specific type and attribute.

        Args:
            device_type (str): The type of the device.
            attribute (str): The attribute to check.
            value (str): The value of the attribute.
        """
        devices = self.get_all_devices()
        return [
            device
            for device in devices
            if device["entity_id"].startswith(device_type) and device["attributes"][attribute] not in value
        ]

    def turn_on(self, device_id, device_type):
        """Turn on a device.

        Args:
            device_id (str): The id of the device.
            device_type (str): The type of the device.
        """
        url = self.host + "/api/services/" + device_type + "/turn_on"
        payload = {"entity_id": device_id}
        response = requests.post(
            url, data=json.dumps(payload), headers=self.headers, timeout=self.timeout, verify=self.verify_ssl
        )
        try:
            response.raise_for_status()
            return response.json()
        except requests.exceptions.RequestException:
            LOG.exception("Error turning on device")
            return None

    def turn_off(self, device_id, device_type):
        """Turn off a device.

        Args:
            device_id (str): The id of the device.
            device_type (str): The type of the device.
        """
        url = self.host + "/api/services/" + device_type + "/turn_off"
        payload = {"entity_id": device_id}
        response = requests.post(
            url, data=json.dumps(payload), headers=self.headers, timeout=self.timeout, verify=self.verify_ssl
        )
        try:
            response.raise_for_status()
            return response.json()
        except requests.exceptions.RequestException:
            LOG.exception("Error turning off device")
            return None

    def call_function(self, device_id, device_type, function, arguments=None):
        """Call a function on a device.

        Args:
            device_id (str): The id of the device.
            device_type (str): The type of the device.
            function (str): The function to call.
            arguments (dict): The arguments to pass to the function.
        """
        url = self.host + "/api/services/" + device_type + "/" + function
        payload = {"entity_id": device_id}
        if arguments:
            for key, value in arguments.items():
                payload[key] = value

        response = requests.post(
            url, data=json.dumps(payload), headers=self.headers, timeout=self.timeout, verify=self.verify_ssl
        )

        try:
            response.raise_for_status()
            return response.json()
        except requests.exceptions.RequestException:
            LOG.exception("Error calling function")
            return None

    def send_assist_command(self, command, arguments=None):
        """Send a command to the Home Assistant Assist websocket endpoint.

        Args:
            command (string): Spoken command to send to Home Assistant.
            arguments (dict, optional): Additional arguments to send. HA currently only supports 'language'
        """
        arguments = arguments or {}
        url = self.host + "/api/conversation/process"
        payload = {
            "text": command,
            "language": arguments.get("language", "en"),
        }
        response = requests.post(
            url, data=json.dumps(payload), headers=self.headers, timeout=self.timeout, verify=self.verify_ssl
        )
        try:
            response.raise_for_status()
            return response.json()
        except requests.exceptions.RequestException:
            LOG.exception("Error sending Assist command")
            return None
